import numpy as np
# try import torch or warn
try:
    import torch
except ImportError:
    print("Torch not found, please install PyTorch to use this decoder.")
    torch = None

from falcon_challenge.config import FalconConfig, FalconTask
from falcon_challenge.interface import BCIDecoder


class MyConfig:
    model_path: str = ""

class SimpleRNNDecoder(BCIDecoder):
    def __init__(self, task_config: FalconConfig, decoder_cfg: MyConfig):
        self._task_config = task_config
        self.dnn = torch.load(decoder_cfg.model_path)

    def reset(self, dataset_tag: str = ""):
        self.neural_history = np.zeros((0, self._task_config.n_channels))

    def predict(self, neural_observations: np.ndarray):
        self.neural_history = np.concatenate([self.neural_history, neural_observations[np.newaxis, :]], axis=0)
        return self.dnn(self.neural_history)



r"""
    Sample NDT2 decoder for the Falcon Challenge.

    H1: https://wandb.ai/joelye9/context_general_bci/runs/edf4h5ym
    M1: https://wandb.ai/joelye9/context_general_bci/runs/93snpffp
"""

import argparse

from falcon_challenge.config import FalconConfig, FalconTask
from falcon_challenge.evaluator import FalconEvaluator

from context_general_bci.falcon_decoder import NDT2Decoder

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--evaluation", type=str, required=True, choices=["local", "remote"]
    )
    parser.add_argument(
        "--model-path", type=str, default='./local_data/ndt2_h1_sample.pth'
    )
    parser.add_argument(
        "--config-stem", type=str, default='falcon/h1/h1',
        help="Name in context-general-bci codebase for config. \
            Currently, directly referencing e.g. a local yaml is not implemented unclear how to get Hydra to find it in search path."
    )
    parser.add_argument(
        "--zscore-path", type=str, default='./local_data/ndt2_zscore_h1.pt'
    )
    parser.add_argument(
        '--split', type=str, choices=['h1', 'h2', 'm1', 'm2'], default='h1',
    )
    parser.add_argument(
        '--phase', choices=['minival', 'test'], default='minival'
    )
    parser.add_argument(
        '--batch-size', type=int, default=1
    )
    parser.add_argument(
        '--continual', action='store_true', default=False # deprecated
    )
    
    args = parser.parse_args()


    evaluator = FalconEvaluator(
        eval_remote=args.evaluation == "remote",
        split=args.split,
        # continual=args.continual
    )

    task = getattr(FalconTask, args.split)
    config = FalconConfig(task=task)
    max_bins = 50 if task in [FalconTask.m1, FalconTask.m2] else 200 # h1
    
    decoder = NDT2Decoder(
        task_config=config,
        model_ckpt_path=args.model_path,
        model_cfg_stem=args.config_stem,
        zscore_path=args.zscore_path,
        max_bins=max_bins,
        dataset_handles=[x.stem for x in evaluator.get_eval_files(phase=args.phase)],
        batch_size=args.batch_size
    )


    evaluator.evaluate(decoder, phase=args.phase)


if __name__ == "__main__":
    main()


from __future__ import annotations

import enum
from typing import Union
from pathlib import Path
import datetime
from dataclasses import dataclass, field

from hydra.core.config_store import ConfigStore
from hydra.core.config_search_path import ConfigSearchPath
from hydra.plugins.search_path_plugin import SearchPathPlugin

H1_NEW_TO_OLD = {
    '19250101T111740': 'S0_set_1',
    '19250101T112404': 'S0_set_2',
    '19250108T110520': 'S1_set_1',
    '19250108T111022': 'S1_set_2',
    '19250108T111455': 'S1_set_3',
    '19250113T120811': 'S2_set_1',
    '19250113T121303': 'S2_set_2',
    '19250115T110633': 'S3_set_1',
    '19250115T111328': 'S3_set_2',
    '19250119T113543': 'S4_set_1',
    '19250119T114045': 'S4_set_2',
    '19250120T115044': 'S5_set_1',
    '19250120T115537': 'S5_set_2',
    '19250126T113454': 'S6_set_1',
    '19250126T114029': 'S6_set_2',
    '19250127T120333': 'S7_set_1',
    '19250127T120826': 'S7_set_2',
    '19250129T112555': 'S8_set_1',
    '19250129T113059': 'S8_set_2',
    '19250202T113958': 'S9_set_1',
    '19250202T114452': 'S9_set_2',
    '19250203T113515': 'S10_set_1',
    '19250203T114018': 'S10_set_2',
    '19250206T112219': 'S11_set_1',
    '19250206T112712': 'S11_set_2',
    '19250209T111826': 'S12_set_1',
    '19250209T112327': 'S12_set_2',
}

class FalconTask(enum.Enum):
    r"""
        Enumerates the tasks that the Falcon challenge supports.
    """
    h1 = "falcon_h1_7d"
    h2 = "falcon_h2_writing"
    m1 = "falcon_m1_finger"
    m2 = "falcon_m2_reach"

@dataclass
class FalconConfig:
    r"""
        User is responsible for copying this appropriately,
        since ultimately these values are to help user to inform decoder what outputs are expected.
    """
    # falcon_h1, falcon_h2_writing, falcon_m1_finger, falcon_m2_reach
    task: FalconTask = FalconTask.h1
    # n_channels: int = 176
    bin_size_ms: int = 20
    # dataset_handles: list[str] = field(default_factory=lambda: []) # Compute with evaluator.get_eval_handles

    @property
    def n_channels(self):
        if self.task == FalconTask.h1:
            return 176
        elif self.task == FalconTask.h2:
            return 192
        elif self.task == FalconTask.m1:
            return 64
        elif self.task == FalconTask.m2:
            return 96
        raise NotImplementedError(f"Task {self.task} not implemented.")

    @property
    def out_dim(self):
        if self.task == FalconTask.h1:
            return 7
        elif self.task == FalconTask.h2:
            return 28
        elif self.task == FalconTask.m1:
            return 16
        elif self.task == FalconTask.m2:
            return 2
        raise NotImplementedError(f"Task {self.task} not implemented.")
        
    def hash_dataset(self, handle: Union[str, Path]):
        r"""
            handle - path.stem of a datafile.
            Convenience function to help identify what "session" a datafile belongs to.. If multiple files per session in real-world time, this may _not_ uniquely identify runfile.
        """
        if isinstance(handle, Path):
            handle = handle.stem
        if self.task == FalconTask.h1:
            if 'sub-HumanPitt' in handle:
                date_hash = handle.split('_ses-')[-1]
                return H1_NEW_TO_OLD[date_hash]
            handle = handle.replace('-', '_')
            handle = '_'.join(handle.split('_')[:-1]) # exclude split annotation
            # print(handle)
            return handle
            # dandi-like atm but not quite determined; e.g. S0_set_1_calib
            # remove split and set information
            # pieces = handle.split('_')
            # for piece in pieces:
                # if piece[0].lower() == 's' and piece != 'set':
                    # return piece
            raise ValueError(f"Could not find session in {handle}.")
        elif self.task == FalconTask.h2:
            return handle.split('_')[1]
        elif self.task == FalconTask.m1: # return date
            # sub-MonkeyL-held-in-minival_ses-20120924_behavior+ecephys.nwb
            # or L_20120924_held_in_eval.nwb
            if 'behavior+ecephys' in handle:
                return handle.split('_')[-2].split('-')[-1]
            return handle.split('_')[1]
        elif self.task == FalconTask.m2:
            if 'behavior+ecephys' in handle: # public sub-MonkeyN-held-in-calib_ses-2020-10-19-Run1_behavior+ecephys.nwb
                run_str = handle.split('_')[1][-4:]
                date_str = ''.join(handle.split('_')[1].split('-')[1:4])
            # sub-MonkeyNRun1_20201019_held_in_eval.nwb 
            else:
                run_str = handle.split('_')[0][-4:]
                date_str = handle.split('_')[1][:8]
            return f'{run_str}_{date_str}'
            

cs = ConfigStore.instance()
cs.store(name="falcon_config", node=FalconConfig)


from typing import List
from functools import partial
import os
import pickle
import torch
import re
from collections import defaultdict
import logging
import numpy as np
from pathlib import Path
from tqdm import tqdm
from sklearn.metrics import r2_score
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset, DataLoader
from edit_distance import SequenceMatcher
from time import time, sleep

from falcon_challenge.config import FalconTask, FalconConfig
from falcon_challenge.interface import BCIDecoder
from falcon_challenge.dataloaders import load_nwb

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Data hash map, what participant submits, keys of comparison dict
DATASET_HELDINOUT_MAP = {
    'h1': {
        'held_in': [
            'S0_set_1', 
            'S0_set_2', 
            'S1_set_1', 
            'S1_set_2', 
            'S1_set_3',
            'S2_set_1',
            'S2_set_2',
            'S3_set_1',
            'S3_set_2',
            'S4_set_1',
            'S4_set_2',
            'S5_set_1',
            'S5_set_2',
        ],
        'held_out': [
            'S6_set_1', 
            'S6_set_2', 
            'S7_set_1', 
            'S7_set_2', 
            'S8_set_1', 
            'S8_set_2', 
            'S9_set_1', 
            'S9_set_2', 
            'S10_set_1', 
            'S10_set_2', 
            'S11_set_1', 
            'S11_set_2', 
            'S12_set_1',
            'S12_set_2',
        ],
    },
    'm1': {
        'held_in': ['20120924', '20120926', '20120927', '20120928'],
        'held_out': ['20121004', '20121017', '20121024'],
    },
    'h2': {
        'held_in': [
            '2022.05.18', 
            '2022.05.23', 
            '2022.05.25', 
            '2022.06.01', 
            '2022.06.03', 
            '2022.06.06', 
            '2022.06.08', 
            '2022.06.13', 
            '2022.06.15', 
            '2022.06.22', 
            '2022.09.01', 
            '2022.09.29', 
            '2022.10.06',
            '2022.10.18', 
            '2022.10.25', 
            '2022.10.27', 
            '2022.11.01', 
            '2022.11.03', 
            '2022.12.08',
            '2022.12.15', 
            '2023.02.28',
        ],
        'held_out': [
            '2023.04.17', 
            '2023.05.31', 
            '2023.06.28', 
            '2023.08.16', 
            '2023.10.09'
        ]
    },
    'm2': {
        'held_in': ['Run1_20201019', 'Run2_20201019', 'Run1_20201020', 'Run2_20201020', 'Run1_20201027', 'Run2_20201027', 'Run1_20201028'],
        'held_out': ['Run1_20201030', 'Run2_20201030', 'Run1_20201118', 'Run1_20201119', 'Run1_20201124', 'Run2_20201124'],
    },
}

# Used to label test server data file names to look for
# These act as a _reduction_ set of the full list of datafiles considered, to specific sessions. Relevant for H1/M2.
def reduce_key(key):
    if key.startswith('Run'):
        return key.split('_')[1]
    if key.startswith('L_'):
        return key
    if key.startswith('S'):
        return key.split('_')[0]    
    return key

HELD_IN_KEYS = {
    FalconTask.h1: ['S0_', 'S1_', 'S2_', 'S3_', 'S4_', 'S5_'],
    FalconTask.m1: ['L_20120924', 'L_20120926', 'L_20120927', 'L_20120928'],
    FalconTask.m2: ['20201019', '20201020', '20201027', '20201028'],
    FalconTask.h2: DATASET_HELDINOUT_MAP['h2']['held_in'],
}

HELD_OUT_KEYS = {
    FalconTask.h1: ['S6_', 'S7_', 'S8_', 'S9_', 'S10_', 'S11_', 'S12_'],
    FalconTask.m1: ['L_20121004', 'L_20121017', 'L_20121024'],
    FalconTask.m2: ['20201030', '20201118', '20201119', '20201124'],
    FalconTask.h2: DATASET_HELDINOUT_MAP['h2']['held_out'],
}

RECOMMENDED_BATCH_SIZES = {
    FalconTask.h1: 8,
    FalconTask.m1: 4,
    FalconTask.h2: 1,
    FalconTask.m2: 7, # max
}

# Development time flag. False allows direct evaluation without payload writing, only usable for local minival.
# Should be set to true for remote evaluation.
# USE_PKLS = False
USE_PKLS = True

BIN_SIZE = 0.02

HELDIN_OR_OUT_MAP = {
    'held_in': "Held In",
    'held_out': "Held Out",
}

def evaluate(
    test_annotation_file: str, # The annotation file for the phase
    user_submission_file: str, # * JY: This appears to always be /submission/submission.csv on EvalAI. No matter - load it as a pickle.
    phase_codename: str, # e.g. minival or test
    **kwargs
):
    r"""
        Test struct:
        {
            'h1': {
                'hash': {
                    'data': tgt,
                    'mask': mask,
                }
            }
        }
        User submission struct:
        {
            'h1': {
                'hash': pred,

                'normalized_latency': 1,
            }
        }
        Evaluate payloads with potentially multiple splits worth of data
        - Low pri: can I provide all results or just one split's worth entry? Currently providing 1, examples just provide 1, but in general would be nice to provide all. User shouldn't be able to submit more than 1, though.
    """
    # ! Want: Locally, test_annotation should be somewhere safe (tmp)
    # ! Remotely, it shoudl be /submission/submission.csv exactly.
    # Ignore explicit annotations provided and directly search for concatenated answers
    logger.info(f"Evaluation: Docker side")
    # test_annotation_file = '/dataset/evaluation_data/answer_key/minival.pkl'
    logger.info(f"Loading GT from {test_annotation_file}")
    logger.info(f"Loading submission from {user_submission_file}")
    logger.info(f"Phase: {phase_codename}")

    result = []
    # Load pickles
    try:
        with open(test_annotation_file, 'rb') as test_annotation_file, open(user_submission_file, 'rb') as user_submission_file:
            test_annotations = pickle.load(test_annotation_file)
            user_submission = pickle.load(user_submission_file)
    except Exception as e:
        logger.error(f"Checking root: {os.listdir('/')}")
        raise ValueError(f"Failed to load submission pickles: {e}. dir is {os.getcwd()}; contents {os.listdir()}.")
    for datasplit in user_submission: # datasplit e.g. h1, m1
        if datasplit not in test_annotations:
            raise ValueError(f"Missing {datasplit} in GT labels.")
        split_annotations = test_annotations[datasplit]
        split_result = {}
        split_result["Normalized Latency"] = user_submission[datasplit]["normalized_latency"]
        del user_submission[datasplit]["normalized_latency"]
        pred_dict = defaultdict(list)
        tgt_dict = defaultdict(list)
        mask_dict = defaultdict(list)
        if 'h2' not in datasplit:
            dset_len_dict = defaultdict(lambda: defaultdict(list))
        if 'm2' in datasplit:
            grouped = {}
            for set_name in user_submission[datasplit]:
                run, date = set_name.split('_')
                if date not in grouped:
                    grouped[date] = []
                grouped[date].append(set_name)
            for date in grouped:
                grouped[date].sort()
            all_datasets = [run for runs in grouped.values() for run in runs]
        else:
            all_datasets = user_submission[datasplit]
        for dataset in all_datasets:
            dataset_pred = user_submission[datasplit][dataset]
            dataset_tgt = split_annotations[dataset]['data']
            dataset_mask = split_annotations[dataset]['mask']
            dataset_pred = dataset_pred[:dataset_mask.shape[0]] # In case excess timesteps are predicted due to batching, reduce
            if dataset in DATASET_HELDINOUT_MAP[datasplit]['held_in']:
                if 'h2' not in datasplit:
                    # For splits with multiple datasets per session (H1 and M2), we need to map predictions, targets, and masks for each dataset to the session ID 
                    session_id = reduce_key(dataset)
                    dset_len_dict['held_in'][session_id].append(dataset_mask.shape[0])
                pred_dict['held_in'].append(dataset_pred)
                tgt_dict['held_in'].append(dataset_tgt)
                mask_dict['held_in'].append(dataset_mask)
            elif dataset in DATASET_HELDINOUT_MAP[datasplit]['held_out']:
                if not 'h2' in datasplit:
                    # For splits with multiple datasets per session (H1 and M2), we need to map predictions, targets, and masks for each dataset to the session ID 
                    session_id = reduce_key(dataset)
                    dset_len_dict['held_out'][session_id].append(dataset_mask.shape[0])
                pred_dict['held_out'].append(dataset_pred)
                tgt_dict['held_out'].append(dataset_tgt)
                mask_dict['held_out'].append(dataset_mask)
            else:
                raise ValueError(f"Dataset {dataset} submitted but not found in held-in or held-out list of split {datasplit}.")
            
        for in_or_out in pred_dict:
            if len(pred_dict[in_or_out]) < len(DATASET_HELDINOUT_MAP[datasplit][in_or_out]):
                raise ValueError(f"Missing predictions for {datasplit} {in_or_out}. User submitted: {user_submission[datasplit].keys()}. Expecting more like: {HELDIN_OR_OUT_MAP[datasplit][in_or_out]}.")
            pred = np.concatenate(pred_dict[in_or_out])
            if 'h2' in datasplit:
                tgt = [y for x in tgt_dict[in_or_out] for y in x]
            else:
                tgt = np.concatenate(tgt_dict[in_or_out])
                dset_lens = dset_len_dict[in_or_out]
            mask = np.concatenate(mask_dict[in_or_out])
            try:
                metrics = FalconEvaluator.compute_metrics_edit_distance(pred, tgt, mask) if 'h2' in datasplit else FalconEvaluator.compute_metrics_regression(pred, tgt, mask, dset_lens)
            except Exception as e:
                raise ValueError(f"Failed to compute metrics for {datasplit} {in_or_out}: {e}. Lengths submitted: {[len(piece) for piece in pred_dict[in_or_out]]}")
            for k in metrics:
                split_result[f'{HELDIN_OR_OUT_MAP[in_or_out]} {k}'] = metrics[k]
        result.append({f'{phase_codename}_split_{datasplit}': split_result})

    print(f"Returning result from phase: {phase_codename}: {result}")
    # Out struct according to https://evalai.readthedocs.io/en/latest/evaluation_scripts.html
    return {"result": result, 'submission_result': result[0]}

class EvalDataset(Dataset):
    r"""
        Simple class to help with batched evaluation.
        data: List of numpy arrays, one per datafile
        datafiles: str names, for cuing models which dataset is being evaluated
        trial_mode: serve trialized data (padded) or just parallelize among datafiles
    """
    def __init__(self, 
                 data: List[np.ndarray], 
                 trial_change: List[np.ndarray], 
                 targets: List[np.ndarray], 
                 eval_mask: List[np.ndarray], 
                 datafiles: List[str],
                 trial_mode=False):
        self.data = data
        self.trial_change = trial_change
        self.targets = targets
        self.eval_mask = eval_mask
        self.datafiles = datafiles
        assert not trial_mode, "Trial mode not implemented."

    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        return (
            self.data[idx], 
            self.targets[idx], 
            self.trial_change[idx], 
            self.eval_mask[idx], 
            idx # needed to index for datafiles which are not natively loaded for pytorch dataloader
        )
    
    def get_datafile(self, idx):
        return self.datafiles[idx]

def simple_collater(batch, task):
    r"""
        Collates a batch of data, targets, trial_change, eval_mask, and datafile_idx.
        data: List of numpy arrays, one per datafile
        datafiles: str names, for cuing models which dataset is being evaluated
    """
    data, targets, trial_change, eval_mask, datafile_idx = zip(*batch)
    data = pad_sequence([torch.tensor(d) for d in data], batch_first=False).numpy()
    if task == FalconTask.h2:
        targets = pad_sequence([torch.tensor(t) for t in targets[0]], batch_first=True).numpy()
    else:
        targets = pad_sequence([torch.tensor(t) for t in targets], batch_first=False).numpy()
    trial_change = pad_sequence([torch.tensor(t) for t in trial_change], batch_first=False).numpy()
    eval_mask = pad_sequence([torch.tensor(t) for t in eval_mask], batch_first=False).numpy() # Serves as a mask for padding as well
    datafile_idx = np.array(datafile_idx)
    return data, targets, trial_change, eval_mask, datafile_idx

class FalconEvaluator:

    def __init__(self, eval_remote=False, split='h1', verbose=False):
        r"""
            verbose: Print out dataset specific metrics for movement tasks.
        """
        self.eval_remote = eval_remote
        assert split in ['h1', 'h2', 'm1', 'm2'], "Split must be h1, h2, m1, or m2."
        if split in ['h1', 'm1', 'm2']:
            self.continual = True
        else:
            self.continual = False
        self.verbose = verbose
        self.dataset: FalconTask = getattr(FalconTask, split)
        self.cfg = FalconConfig(self.dataset)

    @staticmethod
    def get_eval_handles(is_remote: bool, dataset: FalconTask, phase: str = 'minival'):
        if is_remote: # i.e. definitely docker
            data_dir = os.environ.get("EVAL_DATA_PATH")
        else: # possibly docker or local
            if os.path.exists(f"./data/{dataset.name}"):
                logger.info("Using local data directory.")
                data_dir = "data"
            else:
                data_dir = os.environ.get("EVAL_DATA_PATH") # a local docker eval
        data_dir = Path(data_dir) / dataset.name
        if not data_dir.exists():
            raise FileNotFoundError(f"Evaluation data directory {data_dir} not found.")
        if phase == 'test': # TODO wire wherever test is actually stored on remote
            eval_dir = data_dir / f"eval"
        else:
            eval_dir = data_dir / "minival"
        if not eval_dir.exists():
            raise FileNotFoundError(f"Evaluation directory {eval_dir} found but requested phase {phase} not found.")
        return sorted(list(eval_dir.glob("*val*.nwb")))

    def get_eval_files(self, phase: str = 'minival'):
        logger.info("Searching for evaluation data.")
        handles = self.get_eval_handles(self.eval_remote, self.dataset, phase=phase)
        logger.info(f"Found {len(handles)} files.")
        if len(handles) == 0:
            raise FileNotFoundError(f"No files found in {self.dataset.name} for phase {phase}. Note test phase data is only available on EvalAI remote.")
        return handles
    
    def predict_files(self, decoder: BCIDecoder, eval_files: List):
        # returns triple dict, keyed by datafile hash and contains preds, targets, and eval_mask respective. also returns lists compute_time and neural_time
        # TODO this does not return uniquely identifiable data if eval_files is partial, e.g. if we only has set 2 of a day with 2 sets, we'll happily just provide partial predictions.
        # Pre-loop before starting batch loads
        file_neural_data = []
        file_trial_change = []
        file_targets = []
        file_eval_mask = []
        all_neural_times = []
        datafiles = list(sorted(eval_files))
        for datafile in datafiles:
            if not datafile.exists():
                raise FileNotFoundError(f"File {datafile} not found.")
            neural_data, decoding_targets, trial_change, eval_mask = load_nwb(datafile, dataset=self.dataset)
            file_neural_data.append(neural_data)
            file_trial_change.append(trial_change)
            file_targets.append(decoding_targets)
            file_eval_mask.append(eval_mask)
            all_neural_times.append(neural_data.shape[0] * BIN_SIZE)
    
        dataset = EvalDataset(
            data=file_neural_data,
            trial_change=file_trial_change,
            targets=file_targets,
            eval_mask=file_eval_mask,
            datafiles=datafiles,
        )
        
        all_preds = defaultdict(list)
        all_targets = defaultdict(list)
        all_eval_mask = defaultdict(list)
        all_compute_times = []
        
        num_workers = 8
        simple_collater_partial = partial(simple_collater, task=self.dataset)
        dataloader = DataLoader(
            dataset, shuffle=False,
            batch_size=decoder.batch_size,
            num_workers=num_workers,
            persistent_workers=num_workers > 0,
            collate_fn=simple_collater_partial,
        )
        # from time import time
        # for neural_data, decoding_targets, trial_change, eval_mask, datafile in tqdm(dataset):
        for neural_data, decoding_targets, trial_change, eval_mask, datafile_idx in tqdm(dataloader):
            neural_data: np.ndarray
            decoding_targets: np.ndarray
            trial_change: np.ndarray
            eval_mask: np.ndarray
            datafile_idx: np.ndarray
            decoder.reset(dataset_tags=[dataset.datafiles[idx] for idx in datafile_idx])
            trial_preds = []
            # loop_times = []
            # breakpoint()
            for neural_observations, trial_delta_obs, step_mask in zip(neural_data, trial_change, eval_mask):
                neural_observations: np.ndarray
                trial_delta_obs: np.ndarray
                step_mask: np.ndarray
                if self.dataset == FalconTask.h2:
                    assert neural_data.shape[1] == 1, "H2 expects batch size 1."
                    if step_mask[0]:
                        start_time = time()
                        decoder.predict(neural_observations)
                        end_time = time()
                        all_compute_times.append(end_time - start_time)
                    if trial_delta_obs[0]:
                        trial_preds.append(decoder.on_done(trial_delta_obs))
                else:
                    if not self.continual:
                        decoder.on_done(trial_delta_obs)
                    start_time = time()
                    step_prediction = decoder.predict(neural_observations)
                    end_time = time()
                    all_compute_times.append(end_time - start_time)
                    assert step_prediction.shape[1] == self.cfg.out_dim, f"Prediction shape mismatch: {step_prediction.shape[1]} vs {self.cfg.out_dim}."
                    trial_preds.append(step_prediction)
                    
                    # if step_mask.any():
                    #     trial_preds.append(decoder.predict(neural_observations))
                    # else:
                    #     decoder.observe(neural_observations)
                    #     trial_preds.append(np.full((decoder.batch_size, self.cfg.out_dim), np.nan))
                # loop_times.append(time() - loop_start)
            # loop_times = np.array(loop_times)
            # print(f"Loop {len(loop_times)}: {loop_times.mean()} +/- {loop_times.std()}")
            
            if self.dataset == FalconTask.h2:
                datafile_hash = self.cfg.hash_dataset(dataset.get_datafile(datafile_idx[0]))
                all_preds[datafile_hash].append(trial_preds)
                all_targets[datafile_hash].append(decoding_targets)
                all_eval_mask[datafile_hash].append(eval_mask)
            else:
                trial_preds = np.stack(trial_preds) # -> T x B x H
                for idx in range(len(datafile_idx)):
                    datafile_hash = self.cfg.hash_dataset(dataset.get_datafile(datafile_idx[idx]))
                    all_preds[datafile_hash].append(trial_preds[:, idx])
                    all_targets[datafile_hash].append(decoding_targets[:, idx])
                    all_eval_mask[datafile_hash].append(eval_mask[:, idx])
        for k in all_preds:
            if self.dataset == FalconTask.h2:
                all_preds[k] = [x for y in all_preds[k] for x in y]
                all_targets[k] = [x for y in all_targets[k] for x in y]
            else:
                all_preds[k] = np.concatenate(all_preds[k])
                all_targets[k] = np.concatenate(all_targets[k])
            all_eval_mask[k] = np.concatenate(all_eval_mask[k])
        return all_preds, all_targets, all_eval_mask, all_compute_times, all_neural_times

    def evaluate_files(self, decoder: BCIDecoder, eval_files: List):
        all_preds, all_targets, all_eval_mask = self.predict_files(decoder, eval_files)
        metrics = self.compute_metrics(all_preds, all_targets, all_eval_mask)
        return metrics

    def evaluate(
            self, 
            decoder: BCIDecoder, 
            phase: str, 
            held_out_only: bool = False, 
            specific_keys: List = []
        ):
        r"""
            Note: Locally, this can produce metrics, but locally and remotely it should also write a submission file 
            that the actual evaluator on remote uses. The evaluation is done separately on remote.

            held_out_only: Only run predictions on held out
            specific_keys: Overrides held_out_only. Only run predictions on datafiles with specific keys.
        """
        assert phase in ['minival', 'test'], "Phase must be minival or test."
        if phase == 'minival' and (held_out_only or specific_keys):
            logger.warning("Ignoring held_out_only and specific_keys for minival phase.")
            held_out_only = False
        if decoder.batch_size > RECOMMENDED_BATCH_SIZES[self.dataset]:
            logger.warning(f"Decoder batch size {decoder.batch_size} is larger than limit {RECOMMENDED_BATCH_SIZES[self.dataset]} for {self.dataset}, clipping down.")
            decoder.set_batch_size(RECOMMENDED_BATCH_SIZES[self.dataset])
        
        np.random.seed(0)
        # ! TODO ideally seed other libraries as well...? Is that our responsibility?
        
        eval_files = self.get_eval_files(phase=phase)
        metrics = {}
        prediction_env_var = "PREDICTION_PATH" if self.eval_remote else "PREDICTION_PATH_LOCAL"
        prediction_path = os.environ.get(prediction_env_var, './local_prediction.pkl')
        if not prediction_path:
            raise ValueError("PREDICTION_PATH not set in remote env which expects it. Cannot forward to separate evaluate runscript.")
        gt_path = os.environ.get("GT_PATH", './local_gt.pkl')
        if not gt_path:
            raise ValueError("GT_PATH not set in remote env which expects it. Cannot forward to separate evaluate runscript.")
            
        if phase == 'test':
            eval_files_held_in = [f for f in eval_files if any(k in f.name for k in HELD_IN_KEYS[self.dataset])]
            eval_files_held_out = [f for f in eval_files if any(k in f.name for k in HELD_OUT_KEYS[self.dataset])]
            assert len(eval_files) == len(eval_files_held_in) + len(eval_files_held_out), f"Mismatch in extracted eval #: Eval file state is not consistent with benchmark creation settings. Found {len(eval_files)} files, {len(eval_files_held_in)} held in, {len(eval_files_held_out)} held out."

            if specific_keys:
                raise NotImplementedError("not sure what metrics to compute for specific keys yet.")
            elif held_out_only:
                eval_files_held_in = []
            all_preds, all_targets, all_eval_mask, all_compute_times, all_neural_times = self.predict_files(decoder, eval_files_held_out)
            if eval_files_held_in:
                all_preds_held_in, all_targets_held_in, all_eval_mask_held_in, _, _ = self.predict_files(decoder, eval_files_held_in)
                all_preds.update(all_preds_held_in)
                all_targets.update(all_targets_held_in)
                all_eval_mask.update(all_eval_mask_held_in)

        else:
            all_preds, all_targets, all_eval_mask, all_compute_times, all_neural_times = self.predict_files(decoder, eval_files)
        # Indirect remote setup to satisfy EvalAI interface. Save metrics / comparison to file.
        if USE_PKLS:
            inner_pred = {**all_preds}
            inner_tgt_spoof = { # spoof for local mirror of eval ai path, in reality targets are already compiled on eval ai side.
                k: {
                    'data': all_targets[k][all_eval_mask[k]] if self.dataset != FalconTask.h2 else all_targets[k],
                    'mask': all_eval_mask[k],
                } for k in all_targets
            }
            inner_pred['normalized_latency'] = np.sum(all_compute_times) / np.sum(all_neural_times)
            pred_payload = {self.dataset.name: inner_pred}
            truth_payload = {self.dataset.name: inner_tgt_spoof}
        else:
            pass
            
        if USE_PKLS:
            Path(prediction_path).parent.mkdir(parents=True, exist_ok=True)
            with open(prediction_path, 'wb') as f:
                pickle.dump(pred_payload, f)
            Path(gt_path).parent.mkdir(parents=True, exist_ok=True)
            with open(gt_path, 'wb') as f:
                pickle.dump(truth_payload, f)

            if self.eval_remote:
                print("Sleeping before exiting for remote eval - feel free to interrupt for local eval.", flush=True)
                # Gunjan, EvalAI contact says that current static code eval has an issue where the submission dump is only polled by the EvalAI worker comparison script every 5 minutes
                # Sleep so it's definitely available
                sleep(300) 
            else:
                return evaluate(
                    test_annotation_file=gt_path,
                    user_submission_file=prediction_path,
                    phase_codename=phase
                )
        else:
            for k, v in metrics.items():
                logger.info("{}: {}".format(k, v))

    @staticmethod
    def compute_metrics_regression(preds, targets, eval_mask, dset_lens, verbose=False): # Verbose drop-in
        dset_lens = np.cumsum([sum(dset_lens[key]) for key in sorted(dset_lens.keys())])
        masked_points = np.cumsum(~eval_mask)
        dset_lens = [0] + [dset_len - masked_points[dset_len - 1] for dset_len in dset_lens]
        # assumes targets are already masked
        preds = preds[eval_mask]
        if not targets.shape[0] == preds.shape[0]:
            raise ValueError(f"Targets and predictions have different lengths: {targets.shape[0]} vs {preds.shape[0]}.")
        r2_scores = [r2_score(targets[dset_lens[i]:dset_lens[i+1]], preds[dset_lens[i]:dset_lens[i+1]], 
                              multioutput='variance_weighted') for i in range(len(dset_lens) - 1)]
        if verbose:
            dsets = sorted(dset_lens.keys())
            print([f'{k}: {r2}' for k, r2 in zip(dsets, r2_scores)])
            preds_dict = {k: preds[dset_lens[i]:dset_lens[i+1]] for i, k in enumerate(dsets)}
            with open('preds.pkl', 'wb') as f:
                pickle.dump(preds_dict, f)
        return {
            "R2 Mean": np.mean(r2_scores),
            "R2 Std.": np.std(r2_scores)
        }


    @staticmethod
    def compute_metrics_edit_distance(preds, targets, eval_mask):
        if len(preds) != len(targets):
            raise ValueError(f"Targets and predictions have different lengths: {len(targets)} vs {len(preds)}.")
        
        def _to_words(s):
            return s.replace(">", " ").replace(r"([~,!?])", r" \1").split(" ")
        
        char_counts = []
        word_counts = []
        char_distances = []
        word_distances = []
        for pred, target in zip(preds, targets):
            target = "".join([chr(c) for c in target if c != 0])
            matcher = SequenceMatcher([c for c in pred], [c for c in target])
            char_distances.append(matcher.distance())
            char_counts.append(len(target))

            pred = _to_words(pred)
            target = _to_words(target)
            matcher = SequenceMatcher(pred, target)
            word_distances.append(matcher.distance())
            word_counts.append(len(target))

        return {
            "WER": np.sum(word_distances) / np.sum(word_counts),
            "CER": np.sum(char_distances) / np.sum(char_counts),
            "WER Std.": np.std(np.array(word_distances) / np.array(word_counts)),
            "CER Std.": np.std(np.array(char_distances) / np.array(char_counts)),
        }

    def compute_metrics(self, all_preds, all_targets, all_eval_mask=None):
        r"""
            all_preds: array of shape (n_timesteps, k_dim)
            all_targets: array of shape (n_timesteps, k_dim)
            all_eval_mask: array of shape (n_timesteps, k_dim). True if we should evaluate this timestep.
        """
        if self.dataset in [FalconTask.h1, FalconTask.m1, FalconTask.m2]:
            metrics = self.compute_metrics_regression(all_preds, all_targets, all_eval_mask, verbose=self.verbose)
        elif self.dataset in [FalconTask.h2]:
            metrics = self.compute_metrics_edit_distance(all_preds, all_targets, all_eval_mask)
        else:
            raise ValueError(f"Unknown dataset {self.dataset}")
        return metrics


r"""
    JY isn't too clear what happens eval-server side on submission but 1 of 2 patterns seem plausible:
    1. Server receives dockerfile and runs it blind.
    - Submitted docker code should call evaluation with config that points to GT data that only lives remotely.
    - This is the pattern that the Habitat challenge uses.
    In order for the image to access the test data, we can mount the data when we run on eval side.
    `docker run -v /path/on/host:/path/in/container your_image_name`
    2. Evaluator
    - Server receives dockerfile, and runs python evaluate.py --code Dockerfile
    And somehow the evaluator will dynamically import the agent code to run it.
"""


from typing import List
from pathlib import Path
import numpy as np
import pandas as pd 
from pynwb import NWBFile, NWBHDF5IO
from pynwb import TimeSeries
from pynwb.behavior import Position
from pynwb.behavior import BehavioralTimeSeries
from pynwb import ProcessingModule

r"""
    FALCON Consts
"""
BIN_SIZE_MS = 20
BIN_SIZE_S = BIN_SIZE_MS / 1000
FEW_SHOT_CALIBRATION_RATIO = 0.2
FEW_SHOT_CALIBRATION_RATIO_M2 = 0.1
EVAL_RATIO = 0.4
EVAL_RATIO_HUMAN = 0.2
SMOKETEST_NUM = 2

def write_to_nwb(nwbfile: NWBFile, fn: Path):
    fn.parent.mkdir(exist_ok=True, parents=True)
    with NWBHDF5IO(str(fn), 'w') as io:
        io.write(nwbfile)

def create_multichannel_timeseries(data_name, chan_names, data, timestamps=None, unit=''):
    """Creates a BehavioralTimeSeries for multi-channel continuous data"""
    ts = BehavioralTimeSeries(name=data_name)
    for i, chan_name in enumerate(chan_names):
        ts.create_timeseries(name=chan_name,
                             data=data[:,i],
                             unit=unit,
                             timestamps=timestamps,
                             comments=f"columns=[{chan_name}]")

    return ts

def apply_filt_to_multi_timeseries(m_ts, filt_func, name, *args, timestamps=None, **kwargs):
    """apply filtering function to each TimeSeries in BehavioralTimeSeries"""
    ts_keys = list(m_ts.time_series.keys())
    filt_m_ts = BehavioralTimeSeries(name=name)
    for ts_key in ts_keys:
        ts = m_ts[ts_key]
        filt_data = filt_func(ts.data, *args, **kwargs)
        if timestamps is None:
            timestamps = ts.timestamps
        filt_m_ts.create_timeseries(name=ts_key,
                                    data=filt_data,
                                    unit=ts.unit,
                                    comments=ts.comments,
                                    timestamps=timestamps)

    return filt_m_ts

def create_continuous_eval_mask_for_trialized_data(t_new: np.ndarray, trial_start_times: np.ndarray, trial_end_times: np.ndarray):
    convert_trial_start_time = pd.to_datetime(trial_start_times, unit='s').round('20ms').values.astype('float64') * 1e-9
    convert_trial_end_time = pd.to_datetime(trial_end_times, unit='s').round('20ms').values.astype('float64') * 1e-9
    eval_mask = np.full(t_new.size, False)
    # now add a mask for getting within-trial periods
    for start, stop in zip(convert_trial_start_time, convert_trial_end_time):
        start_ind = np.searchsorted(t_new, start)
        stop_ind = np.searchsorted(t_new, stop)
        eval_mask[start_ind:stop_ind] = True
    return eval_mask


r"""
    Sample ridge regression decoder for the Falcon Challenge.
"""

import argparse
from falcon_challenge.config import FalconConfig, FalconTask
from falcon_challenge.evaluator import FalconEvaluator

from decoder_demos.sklearn_decoder import SKLearnDecoder

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--evaluation", type=str, required=True, choices=["local", "remote"]
    )
    parser.add_argument(
        "--model-path", type=str, required=False, default='./local_data/sklearn_FalconTask.h1.pkl'
    )
    parser.add_argument(
        '--split', type=str, choices=['h1', 'h2', 'm1', 'm2'], default='h1',
    )
    parser.add_argument(
        '--phase', choices=['minival', 'test'], default='minival'
    )
    parser.add_argument('--batch-size', type=int, help='size of batch for evaluation', default=1)
    args = parser.parse_args()

    task = getattr(FalconTask, args.split)
    config = FalconConfig(
        task=task,
    )

    decoder = SKLearnDecoder(task_config=config, model_path=args.model_path, batch_size=args.batch_size)

    evaluator = FalconEvaluator(
        eval_remote=args.evaluation == "remote",
        split=args.split,
    )
    evaluator.evaluate(decoder, phase=args.phase)


if __name__ == "__main__":
    main()


import glob
import subprocess
import shlex

""" Run script to make datasets from raw files into NWB format """

base_dir = "/snel/share/share/data/rouse/RTG/"
sessions = glob.glob(base_dir + f"L*AD_Unrect_EMG.mat")

for sess in sessions: 
    idx = sess.split('/')[-1].split('_')[0]
    monkey = idx[0]
    sess = idx[1:]
    cmd = f"python m1_reachgrasp_preprocv2.py {sess} {monkey}"
    subprocess.run(shlex.split(cmd))

# base_dir = "/snel/share/share/data/bg2/t5_handwriting/CORP_data/stability_benchmark"
# sessions = glob.glob(base_dir + "/*/*.mat")

# for sess in sessions: 
#     ext = sess.split('/')[-2]
#     sess = '.'.join(sess.split('/')[-1].split('.')[1:-1])
#     cmd = f"python h2_preproc.py {sess} {ext}"
#     subprocess.run(shlex.split(cmd))




from pathlib import Path
import shutil

root = Path('./data/h1_src')
out = Path('./data/h1')

root = Path('./data/m2/preproc_src')
out = Path('./data/m2')
out.mkdir(exist_ok=True)
# Assemble eval
eval_target = out / 'eval'
eval_target.mkdir(exist_ok=True)
eval_files = list(root.glob('held_in/*eval.nwb')) + list(root.glob('held_out/*eval.nwb'))
# Copy
for ef in eval_files:
    shutil.copy(ef, eval_target / ef.name)

# Assemble minival
minival_target = out / 'minival'
minival_target.mkdir(exist_ok=True)
minival_files = list(root.glob('held_in/*minival.nwb'))
# Copy
for mf in minival_files:
    shutil.copy(mf, minival_target / mf.name)

# Assemble calibration
calib_target = out / 'held_in_calib'
calib_target.mkdir(exist_ok=True)
calib_files = list(root.glob('held_in/*calib.nwb'))
# Copy
for cf in calib_files:
    shutil.copy(cf, calib_target / cf.name)

calib_target = out / 'held_out_calib'
calib_target.mkdir(exist_ok=True)
calib_files = list(root.glob('held_out/*calib.nwb'))
# Copy
for cf in calib_files:
    shutil.copy(cf, calib_target / cf.name)


#%%
r"""
    SN:
    FingerAnglesTIMRL: Contains the normalized positions for all five digits in the range [0, 1] (0 is full extension, one is full flexion). There are five columns (i.e. for thumb, index, middle, ring, and small fingers), though only two are relevant here. The position of the index finger is in index 1 (in Python indices), and the positions of the middle, ring, and small fingers (which were always grouped together) are in index 3. These samples are collected at 1kHz resolution
    Channels: Contains the millisecond times at which a spike was detected (times relative to block beginning, see ExperimentTime below) for each channel. Spikes were detected by the Cerebus at a -4.5RMS threshold after a 250Hz high-pass filter on raw 30kHz data. No re-referencing schemes were applied
    NeuralFeature: Contains the 2kHz samples collected in each 1ms cycle absolute-valued and summed. The number of samples within each 1ms is in the SampleWidth field. One could calculate a spike-band power measure by summing across the time dimension in NeuralFeature and dividing by the sum of the same time window in SampleWidth (i.e. to get a mean). 2kHz samples were filtered to the 300-1,000Hz band by the Cerebus before transmission to our task hardware
    ExperimentTime: Contains the values of a counter that increments every 1ms since the beginning of the block.
    TrialSuccess: Indicates whether the trial's targets were acquired successfully (likely to be all 1 since this was physically controlled by the monkey's fingers)
    BlankTrial: Indicates whether the display was provided to the monkey. Any trials where this is set to 1 should be ignored (i.e. the monkey's screen was off)
    TargetPos: The center positions of each 1D target (5D vector per trial). As with FingerAnglesTIMRL, only Python indices 1 and 3 matter here
    TargetScaling: A pseudo-arbitrary number indicating the size of targets. One can compute the width of targets in the same. units as FingerAnglesTIMRL with width = 0.0375 * (1 + TargetScaling / 100)
    Each element of mat['z'][0] corresponds to one trial
    Note that, due to how the system was developed, there are 2ms missing between each trial. The spike data for these was collected if absolutely needed, but would be a pain to fill it in. These absent 2ms are reflected in ExperimentTimes
"""
# 1. Extract raw datapoints
import math
from pathlib import Path
import numpy as np
from datetime import datetime
from pprint import pprint

from scipy.io import loadmat
from scipy.signal import resample_poly
import matplotlib.pyplot as plt

from dateutil.tz import tzlocal

import pynwb
from pynwb import NWBFile, TimeSeries

from decoder_demos.filtering import smooth
from preproc.nwb_create_utils import (
    FEW_SHOT_CALIBRATION_RATIO_M2 as FEW_SHOT_CALIBRATION_RATIO, EVAL_RATIO, SMOKETEST_NUM, BIN_SIZE_MS, BIN_SIZE_S,
    create_multichannel_timeseries,
    write_to_nwb
)

root = Path('./data/m2/raw')
out_root = Path('./data/m2/preproc_src')
files = list(root.glob('*.mat'))
KIN_LABELS = ['index', 'mrs']
CHANNEL_EXPECTATION = 96
TARGET_FINGERS = [1, 3] # Index finger and middle finger group
FS = 1000
pprint(files)

DATA_SPLITS = {
    'Z_Joker_2020-10-19_Run-002': 'held_in',
    'Z_Joker_2020-10-19_Run-003': 'held_in',
    'Z_Joker_2020-10-20_Run-002': 'held_in',
    'Z_Joker_2020-10-20_Run-003': 'held_in',
    'Z_Joker_2020-10-27_Run-002': 'held_in',
    'Z_Joker_2020-10-27_Run-003': 'held_in',
    'Z_Joker_2020-10-28_Run-001': 'held_in',
    'Z_Joker_2020-10-30_Run-001': 'held_out',
    'Z_Joker_2020-10-30_Run-002': 'held_out',
    'Z_Joker_2020-11-18_Run-001': 'held_out',
    'Z_Joker_2020-11-19_Run-001': 'held_out',
    'Z_Joker_2020-11-24_Run-001': 'held_out',
    'Z_Joker_2020-11-24_Run-002': 'held_out'
}

DATA_RUN_SET = {
    'Z_Joker_2020-10-19_Run-002': 1,
    'Z_Joker_2020-10-19_Run-003': 2,
    'Z_Joker_2020-10-20_Run-002': 1,
    'Z_Joker_2020-10-20_Run-003': 2,
    'Z_Joker_2020-10-27_Run-002': 1,
    'Z_Joker_2020-10-27_Run-003': 2,
    'Z_Joker_2020-10-28_Run-001': 1,
    'Z_Joker_2020-10-30_Run-001': 1,
    'Z_Joker_2020-10-30_Run-002': 2,
    'Z_Joker_2020-11-18_Run-001': 1,
    'Z_Joker_2020-11-19_Run-001': 1,
    'Z_Joker_2020-11-24_Run-001': 1,
    'Z_Joker_2020-11-24_Run-002': 2
}
# Note that intertrial spikes are not available
#%%
def create_nwb_shell(path: Path, split, suffix='full', extra_note=""):
    start_date = path.stem.split('_')[2] # YYYY-MM-DD
    session_start_time = datetime.strptime(start_date, '%Y-%m-%d')
    # inject arbitrary hour to distinguish sets
    session_start_time = session_start_time.replace(hour=12 + DATA_RUN_SET[path.stem])
    # hash = '_'.join(path.stem.split('_')[2:]) # includes date and run, remap run
    session_hash = start_date + '_' + f'Run{DATA_RUN_SET[path.stem]}'
    # breakpoint()
    subject = pynwb.file.Subject(
        subject_id=f'MonkeyN-{split}-{suffix}',
        description=f'MonkeyN, Chestek Lab, number indicates experimental set from day.{extra_note}',
        species='Rhesus macaque',
        sex='M',
        age='P8Y',
    )
    file_id = f'MonkeyN_{session_start_time.strftime("%m-%d-%H:%M")}_{split}'
    # file_id = f'MonkeyN_{session_start_time.strftime("%m-%d-%H:%M")}_run{DATA_RUN_SET[path.stem]}_{split}'
    f = NWBFile(
        session_description='M2 data',
        identifier=file_id,
        subject=subject,
        session_start_time=session_start_time,
        experimenter='Samuel R. Nason-Tomaszewski and Matthew J. Mender',
        lab='Chestek Lab',
        institution='University of Michigan',
        experiment_description='Two finger group movement in NHP. Behavior provided in 20ms bins, observation interval at trial ends may include a bit of neural data from partial bin.',
        session_id=session_hash # determines the fn
    )
    device = f.create_device(name='Blackrock Utah Array', description='2x64-channel array, 96 active')
    main_group = f.create_electrode_group(
        name='M1_array',
        description='Hand area 2x64-channel array (96 active channels in recording system)',
        location='M1',
        device=device,
    )
    f.add_trial_column(name="tgt_loc", description="location of target (0-1 AU)")
    f.add_trial_column(name="trial_num", description="trial number as in experiment")
    for i in range(CHANNEL_EXPECTATION):
        f.add_electrode(
            id=i,
            x=np.nan, y=np.nan, z=np.nan,
            imp=np.nan,
            location='M1',
            group=main_group,
            filtering='250Hz HPF, -4.5RMS threshold'
        )
    return f

def filt_single_trial(trial):
    # reduce raw
    assert len(trial['Channel']) == CHANNEL_EXPECTATION
    return {
        'fingers': trial['FingerAnglesTIMRL'][:, TARGET_FINGERS],
        'spikes': trial['Channel'], # spike time to nearest ms
        'time': trial['ExperimentTime'], # Clock time since start of block. 0 on first step i.e. start of bin.
        'target': trial['TargetPos'][TARGET_FINGERS],
        'trial_num': trial['TrialNumber'],
    }

def to_nwb(path: Path, ):
    full_payload = loadmat(str(path), simplify_cells=True)['z']
    # Skip trial 0 - no data, used for block setup
    full_payload = full_payload[1:]
    full_payload = [i for i in full_payload if not i['BlankTrial']] # 1 if screen was off, no trial run
    full_payload = list(map(filt_single_trial, full_payload))
    def create_and_write(
        payload, suffix
    ):
        extra_note = ""
        if path.stem == 'Z_Joker_2020-10-19_Run-002' and suffix in ['calib', 'full', 'in_day_oracle']:
            extra_note = " Dropped trial in this file, see trial dataframe." # Cannot edit description after creation.
        nwbfile = create_nwb_shell(path, DATA_SPLITS[path.stem], suffix=suffix, extra_note=extra_note)
        cont_bhvr = []
        # cont_vel = []
        all_spikes = [[] for _ in range(CHANNEL_EXPECTATION)]
        cont_time = []
        start_time = 0
        # breakpoint()
        for i, trial_data in enumerate(payload):
            time = (trial_data['time'].astype(float) / 1000) # To ms
            if not start_time:
                start_time = time[0]
            time -= start_time
            nwbfile.add_trial(
                start_time=time[0],
                stop_time=time[-1],
                tgt_loc=trial_data['target'],
                trial_num = trial_data['trial_num']
            )

            bhvr = trial_data['fingers']
            
            # Per-trial Downsample to 20ms
            # time = time[::math.ceil(FS * BIN_SIZE_MS / 1000)] # This effectively rounds up
            # EDGE_PAD = 160 # reduce edge ringing, in ms
            # y_padded = np.pad(bhvr, ((EDGE_PAD, EDGE_PAD), (0, 0)), mode='edge',)
            # y_padded = smooth(y_padded, 120, 40) # Sam says Gaussian is fine

            # y_resampled_padded = resample_poly(y_padded, math.ceil(FS / BIN_SIZE_MS), 1000)
            # bhvr = y_resampled_padded[int(EDGE_PAD / BIN_SIZE_MS):int(-EDGE_PAD / BIN_SIZE_MS)]
            cont_bhvr.append(bhvr)
            # all_vel.append(np.gradient(bhvr, axis=0))
            # assert np.isclose(np.diff(time), BIN_SIZE_S).all(), "Expecting timestamps to be about 20ms apart"
            cont_time.append(time)

            for j, spike in enumerate(trial_data['spikes']):
                spike_data = spike['SpikeTimes']
                if isinstance(spike_data, int):
                    spike_data = [spike_data]
                all_spikes[j].extend(spike_data)

        for i, spikes in enumerate(all_spikes):
            nwbfile.add_unit(
                id=i,
                spike_times=np.array(spikes) / 1000 - start_time,
                electrodes=[i],
                obs_intervals=[[0., cont_time[-1][-1]]]
                # obs_intervals=[[i[0], i[-1] + BIN_SIZE_S] for i in all_time]
            )
        # trial_diffs = [all_time[i+1][0] - all_time[i][-1] for i in range(len(all_time) - 1)]
        # trial_diff_raw = [payload[i+1]['time'][0] - payload[i]['time'][-1] for i in range(len(payload) - 1)]
        # print(f"Max diff b/n consecutive trials: {max(trial_diffs):.4f}")
        # print(f"Max diff b/n consecutive trials (raw): {max(trial_diff_raw):.4f}")
        # Note there's one long drop in `20201019`
        # if max(trial_diff_raw) > 3:
            # breakpoint()
        # all_time = np.concatenate(all_time, axis=0)
        # diff_check = np.diff(all_time).max()
        # assert (diff_check > 0), "Expecting time to be monotonically increasing across trials."
        # print(f"Max diff b/n consecutive timebins: {diff_check}")
        
        # Form continuous behavior and time
        cat_bhvr = np.concatenate(cont_bhvr, axis=0)
        cat_time = np.concatenate(cont_time, axis=0).round(3)
        assert np.diff(cat_time).min() > 0, "Expecting time to be monotonically increasing."
        interp_time = np.arange(cat_time[0], cat_time[-1], 0.001)
        cat_mask = np.zeros_like(interp_time, dtype=bool)
        for interval in cont_time:
            cat_mask[(interp_time >= interval[0]) & (interp_time <= interval[-1])] = True
        interp_bhvrs = [np.interp(interp_time, cat_time, y) for y in cat_bhvr.T]
        interp_bhvr = np.stack(interp_bhvrs, axis=1)
        # smooth and downsample
        EDGE_PAD = 160 # reduce edge ringing, in ms
        downsample_ratio = math.ceil(FS * BIN_SIZE_MS / 1000)
        def get_reshape(vec: np.ndarray, downsample_ratio: int):
            if len(vec) % downsample_ratio:
                vec = vec[:-(len(vec) % downsample_ratio)]
            vec = vec.reshape(-1, downsample_ratio)
            return vec
        downsampled_time = get_reshape(interp_time, downsample_ratio)[:, -1] # Get RHS of bin
        pad_bhvr = np.pad(interp_bhvr, ((EDGE_PAD, EDGE_PAD), (0, 0)), mode='edge',)
        smth_bhvr = smooth(pad_bhvr, 120, 40) # Sam says Gaussian is fine
        resampled_bhvr = resample_poly(smth_bhvr, math.ceil(FS / BIN_SIZE_MS), 1000)
        downsampled_bhvr = resampled_bhvr[int(EDGE_PAD / BIN_SIZE_MS):int(-EDGE_PAD / BIN_SIZE_MS)]
        if smth_bhvr.shape[0] % downsample_ratio:
            downsampled_bhvr = downsampled_bhvr[1:] # Right bias on extra bin from resample
        downsampled_vel = np.gradient(downsampled_bhvr, axis=0)
        downsampled_mask = get_reshape(cat_mask, downsample_ratio).mean(axis=1).round().astype(bool)
        # assert downsampled_mask.all(), "Expecting mask to be all true."
        if downsampled_bhvr.shape[0] != downsampled_time.shape[0]:
            breakpoint()
        ts = create_multichannel_timeseries(
            data_name='finger_pos',
            chan_names=KIN_LABELS,
            data=downsampled_bhvr,
            # data=np.concatenate(downsampled_bhvr, axis=0),
            timestamps=downsampled_time, # timestamps denote wall-clock sample is drawn - not evenly spaced
            unit='AU'
        )
        nwbfile.add_acquisition(ts)

        ts = create_multichannel_timeseries(
            data_name='finger_vel',
            chan_names=KIN_LABELS,
            data=downsampled_vel,
            # data=np.concatenate(downsampled_vel, axis=0),
            timestamps=downsampled_time,
            unit='AU'
        )
        nwbfile.add_acquisition(ts)

        nwbfile.add_acquisition(
            TimeSeries(
                name='eval_mask',
                description='Timesteps to keep covariates (for training, eval).',
                timestamps=downsampled_time,
                data=downsampled_mask,
                # data=np.ones_like(downsampled_time, dtype=bool),
                unit='bool'
            )
        )
        date_str = path.stem.split('_')[2].replace('-', '')
        run = DATA_RUN_SET[path.stem]
        new_prefix = f'sub-MonkeyNRun{run}_{date_str}_{DATA_SPLITS[path.stem]}'
        write_to_nwb(nwbfile, out_root / DATA_SPLITS[path.stem] / f"{new_prefix}_{suffix}.nwb")
        print(f"Written ", out_root / DATA_SPLITS[path.stem] / f"{new_prefix}_{suffix}.nwb")
    eval_num = int(len(full_payload) * EVAL_RATIO)
    eval_trials = full_payload[-eval_num:]
    create_and_write(eval_trials, 'eval')
    create_and_write(full_payload, 'full')
    in_day_full_trials = full_payload[:-eval_num]
    if DATA_SPLITS[path.stem] == 'held_in':
        create_and_write(in_day_full_trials, 'calib')

        minival_trials = full_payload[:SMOKETEST_NUM]
        create_and_write(minival_trials, 'minival')
    else:
        calibration_num = math.ceil(len(full_payload) * FEW_SHOT_CALIBRATION_RATIO)
        calibration_trials = full_payload[:calibration_num]
        create_and_write(calibration_trials, 'calib')

        create_and_write(in_day_full_trials, 'in_day_oracle')


for f in files:
    to_nwb(f)
# ! Ok...
# 1. create splits
# 2. create acq downsample
#%%
# Check!
payload = loadmat(str(files[0]), simplify_cells=True)['z']
# Skip trial 0 - no data, used for block setup
payload = payload[1:]
payload = [i for i in payload if not i['BlankTrial']] # 1 if screen was off, no trial run
payload = list(map(filt_single_trial, payload))
from matplotlib import pyplot as plt
trial = 0

time = payload[trial]['time'].astype(float)
fingers = payload[trial]['fingers']

SRC_BINS_PER_TGT = math.ceil(FS * BIN_SIZE_MS / 1000)
downsample_time = time[::SRC_BINS_PER_TGT] # This effectively rounds up
# crop, then bin
if len(fingers) % SRC_BINS_PER_TGT:
    bin_fingers = fingers[:-(len(fingers) % SRC_BINS_PER_TGT)]
    bin_time = time[:-(len(time) % SRC_BINS_PER_TGT)]
bin_time = bin_time.reshape(-1, SRC_BINS_PER_TGT).mean(axis=1)
bin_fingers = bin_fingers.reshape(-1, SRC_BINS_PER_TGT, bin_fingers.shape[-1]).mean(axis=1)
# plt.plot(bin_time, np.gradient(bin_fingers, axis=0))

print(fingers.shape)
# print(downsample_fingers.shape)
EDGE_PAD = 160 # reduce edge ringing, in ms
y_padded = np.pad(fingers, ((EDGE_PAD, EDGE_PAD), (0, 0)), mode='edge',)

# Low pass
print(y_padded.shape)
y_padded = smooth(y_padded, 100, 25)

from scipy.signal import resample_poly
# Resample the padded signal
y_resampled_padded = resample_poly(y_padded, math.ceil(FS / BIN_SIZE_MS), 1000)
y_resampled = y_resampled_padded[int(EDGE_PAD / BIN_SIZE_MS):int(-EDGE_PAD / BIN_SIZE_MS)]
downsample_fingers = y_resampled

# Resampling methods produce edge artifacts
# downsample_fingers = resample_poly(fingers, math.ceil(FS / BIN_SIZE_MS), 1000) # This should round up
# downsample_fingers = resample(fingers, math.ceil(len(fingers) / (FS * BIN_SIZE_MS / 1000)), axis=0) # This should round up
# plt.plot(time, fingers)
# plt.plot(downsample_time, downsample_fingers)
plt.plot(downsample_time, np.gradient(downsample_fingers, axis=0), label='grad')


#%%
from pathlib import Path
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import numpy as np
from pynwb import NWBHDF5IO
from data_demos.styleguide import set_style
from falcon_challenge.dataloaders import bin_units
set_style()

#%% 
# Load the NWB file
test_path = Path('./data/m2/held_out_calib/')
test_files = sorted(test_path.glob('*.nwb'))

#%% 
from falcon_challenge.config import FalconTask
from falcon_challenge.dataloaders import load_nwb
units, cov, time, mask = load_nwb(test_files[0], dataset=FalconTask.m2)
#%% 
from sklearn.metrics import r2_score
from decoder_demos.decoding_utils import (
    zscore_data,
    generate_lagged_matrix,
    apply_neural_behavioral_lag,
    fit_and_eval_decoder
)
from preproc.filtering import apply_exponential_filter

#%% 
n_hist = 8

n_fracs = np.array([0.1, 0.3, 0.5, 0.75, 1.0]) # Going to do 0.5 based on this plot.

trials = np.full((len(test_files), len(n_fracs)), np.nan)
perfs = np.full((len(test_files), len(n_fracs)), np.nan)

for ii in range(len(test_files)): 
    print(f'Processing {test_files[ii]}')
    spikes, emg, time, eval_mask = load_nwb(test_files[ii], dataset=FalconTask.m2)


    # ds_n_trials = len(n_fracs)
    trials[ii, :] = n_fracs * len(eval_mask)

    for jj in range(len(n_fracs)): 
        # print(f'Decoding {ds_n_trials[jj]} trials')
        # ntr = ds_n_trials[jj]
        # last_trial = tinfo.iloc[ntr]
        # end_time = last_trial['end_time']
        # end_ind = np.argmin(np.abs(time - end_time))
        end_ind = int(n_fracs[jj] * len(eval_mask))
        rel_spikes = spikes[:end_ind]
        rel_emg = emg[:end_ind]
        rel_mask = eval_mask[:end_ind]
        rel_ss = apply_exponential_filter(rel_spikes)
        emg_trials = rel_emg[rel_mask]
        ss_trials = rel_ss[rel_mask]
        
        split_ind = int(0.8 * len(emg_trials))
        smoothed_spikes_train, smoothed_spikes_valid = np.split(ss_trials, [split_ind])
        emg_train, emg_valid = np.split(emg_trials, [split_ind])

        train_neural = generate_lagged_matrix(zscore_data(smoothed_spikes_train), n_hist)
        valid_neural = generate_lagged_matrix(zscore_data(smoothed_spikes_valid), n_hist)
        train_behavioral = emg_train[n_hist:]
        valid_behavioral = emg_valid[n_hist:]

        valid_score, _ = fit_and_eval_decoder(
            train_neural,
            train_behavioral,
            valid_neural,
            valid_behavioral,
            return_preds=False
        )

        perfs[ii, jj] = valid_score


#%% 

plt.figure(facecolor='w')
for kk in range(len(test_files)): 
    plt.plot(trials[kk, :], perfs[kk, :], 'o-', label=f'Dataset {kk+1}')
plt.hlines(0.5, 0, 150, color='k', linestyle='--')
plt.ylim(-0.2, 1.2)
plt.xlabel('Timepoints provided (20ms bins)')
# plt.xlabel('Number of trials provided')
plt.ylabel('R2 score')
plt.grid(alpha=0.4)

# %%
plt.figure(facecolor='w')
for kk in range(len(test_files)): 
    plt.plot(trials[kk, :] * 0.2, perfs[kk, :], 'o-', label=f'Dataset {kk+1}')
plt.hlines(0.5, 0, 35, color='k', linestyle='--')
plt.ylim(-0.2, 1.2)
plt.xlabel('Number of valid trials for scoring')
plt.ylabel('R2 score')
plt.grid(alpha=0.4)

# %%



# EvalAI worker stub, dev-testing
import os
import importlib

def main():
    pred_file = os.environ["PREDICTION_PATH"]
    gt_file = os.environ["GT_PATH"]

    challenge_phase = "test"

    CHALLENGE_IMPORT_STRING='falcon_challenge.evaluator'
    challenge_module = importlib.import_module(CHALLENGE_IMPORT_STRING)
    challenge_module.evaluate(
        pred_file,
        gt_file,
        challenge_phase,
        submission_metadata={}
    )

if __name__ == "__main__":
    main()


import numpy as np
from sklearn.linear_model import Ridge
from sklearn.model_selection import GridSearchCV

TRAIN_TEST = (0.8, 0.2)

def zscore_data(data): 
    """
    Z-scores the input data.

    Parameters:
    data (np.ndarray): The input data.

    Returns:
    np.ndarray: The z-scored data.
    """
    m = np.mean(data, axis=0)
    s = np.std(data, axis=0)
    s[s==0] = 1
    return (data - m) / s

def generate_lagged_matrix(input_matrix: np.ndarray, lag: int):
    """
    Generate a lagged version of an input matrix.

    Parameters:
    input_matrix (np.ndarray): The input matrix.
    lag (int): The number of lags to consider.

    Returns:
    np.ndarray: The lagged matrix.
    """
    # Initialize the lagged matrix
    lagged_matrix = np.zeros((input_matrix.shape[0] - lag, input_matrix.shape[1] * (lag + 1)))

    # Fill the lagged matrix
    for i in range(lag + 1):
        lagged_matrix[:, i*input_matrix.shape[1]:(i+1)*input_matrix.shape[1]] = input_matrix[lag-i : (-i if i != 0 else None)]

    return lagged_matrix


def apply_neural_behavioral_lag(neural_matrix: np.ndarray, behavioral_matrix: np.ndarray, lag: int):
    """
    Apply a lag to the neural matrix and the behavioral matrix.

    Parameters:
    neural_matrix (np.ndarray): The neural matrix. (Time x Channels)
    behavioral_matrix (np.ndarray): The behavioral matrix. (Time x Dimensions)
    lag (int): The number of bins to consider.

    Returns:
    np.ndarray: The lagged neural matrix.
    np.ndarray: The lagged behavioral matrix.
    """

    if lag != 0:
        # Apply the lag to the neural matrix
        neural_matrix = neural_matrix[:-lag, :]

        # Apply the lag to the behavioral matrix
        behavioral_matrix = behavioral_matrix[lag:, :]

    return neural_matrix, behavioral_matrix


def fit_and_eval_decoder(
    train_rates: np.ndarray,
    train_behavior: np.ndarray,
    eval_rates: np.ndarray,
    eval_behavior: np.ndarray,
    grid_search: bool=True,
    return_preds: bool=False
):
    """Fits ridge regression on train data passed
    in and evaluates on eval data

    Parameters
    ----------
    train_rates :
        2d array time x units.
    train_behavior :
        2d array time x output dims.
    eval_rates :
        2d array time x units
    eval_behavior :
        2d array time x output dims
    grid_search :
        Whether to perform a cross-validated grid search to find
        the best regularization hyperparameters.

    Returns
    -------
    float
        Uniform average R2 score on eval data
    """
    if np.any(np.isnan(train_behavior)):
        train_rates = train_rates[~np.isnan(train_behavior)[:, 0]]
        train_behavior = train_behavior[~np.isnan(train_behavior)[:, 0]]
    if np.any(np.isnan(eval_behavior)):
        eval_rates = eval_rates[~np.isnan(eval_behavior)[:, 0]]
        eval_behavior = eval_behavior[~np.isnan(eval_behavior)[:, 0]]
    assert not np.any(np.isnan(train_rates)) and not np.any(np.isnan(eval_rates)), \
        "fit_and_eval_decoder: NaNs found in rate predictions within required trial times"

    if grid_search:
        decoder = GridSearchCV(Ridge(), {"alpha": np.logspace(-5, 5, 20)})
    else:
        decoder = Ridge(alpha=1e-2)
    decoder.fit(train_rates, train_behavior)
    if return_preds:
        return decoder.score(eval_rates, eval_behavior), decoder, decoder.predict(eval_rates)
    else:
        return decoder.score(eval_rates, eval_behavior), decoder
